@@ -1,4 +1,9 @@
 mod test_util;
+use evercrypt_sys::evercrypt_bindings::{
+    EverCrypt_AutoConfig2_has_aesni, EverCrypt_AutoConfig2_has_avx,
+    EverCrypt_AutoConfig2_has_movbe, EverCrypt_AutoConfig2_has_pclmulqdq,
+    EverCrypt_AutoConfig2_has_sse,
+};
 use test_util::*;
 
 use evercrypt::aead::{Aead, Error, Mode};
@@ -78,6 +83,13 @@ fn test_wycheproof() {
                 }
                 _ => panic!("Unknown algorithm {:?}", test_vec.algorithm),
             };
+            if !hacl_aes_available()
+                && (algorithm == Mode::Aes128Gcm || algorithm == Mode::Aes256Gcm)
+            {
+                println!("⚠️  AES NOT AVAILABLE ON THIS PLATFORM!");
+                *skipped_tests += testGroup.tests.len();
+                continue;
+            }
             let invalid_iv = if testGroup.ivSize != 96 { true } else { false };
 
             for test in testGroup.tests.iter() {
@@ -102,7 +114,14 @@ fn test_wycheproof() {
                 let exp_tag = hex_str_to_bytes(&test.tag);
                 let key = hex_str_to_bytes(&test.key);
 
-                let cipher = Aead::new(algorithm, &key).unwrap();
+                let cipher = match Aead::new(algorithm, &key) {
+                    Ok(c) => c,
+                    Err(_) => {
+                        println!("⚠️  Skipping {:?} because it's not available.", algorithm);
+                        *skipped_tests += 1;
+                        continue;
+                    }
+                };
                 let (ctxt, tag) = match cipher.encrypt(&msg, &nonce, &aad) {
                     Ok(v) => v,
                     Err(e) => {
@@ -127,6 +146,14 @@ fn test_wycheproof() {
                     ctxt_comb.split_at(ctxt_comb.len() - cipher.tag_size()),
                     (&ctxt[..], &tag[..])
                 );
+                let mut in_place_payload = msg.clone();
+                let tag_in_place = cipher
+                    .encrypt_in_place(&mut in_place_payload, &nonce, &aad)
+                    .unwrap();
+                assert_eq!(
+                    (&in_place_payload[..], &tag_in_place[..]),
+                    (&ctxt[..], &tag[..])
+                );
                 let msg_decrypted = match cipher.decrypt(&ctxt, &tag, &nonce, &aad) {
                     Ok(m) => m,
                     Err(_) => {
@@ -137,6 +164,10 @@ fn test_wycheproof() {
                 assert_eq!(msg, msg_decrypted);
                 let msg_decrypted_comb = cipher.decrypt_combined(&ctxt_comb, &nonce, &aad).unwrap();
                 assert_eq!(msg, msg_decrypted_comb);
+                cipher
+                    .decrypt_in_place(in_place_payload.as_mut_slice(), &tag_in_place, &nonce, &aad)
+                    .unwrap();
+                assert_eq!(msg, in_place_payload);
                 *tests_run += 1;
             }
         }
@@ -155,7 +186,13 @@ fn key_gen_self_test() {
     fn run(algorithm: Mode) {
         let msg = b"Evercrypt rulez";
         let aad = b"associated data";
-        let cipher = Aead::init(algorithm).unwrap();
+        let cipher = match Aead::init(algorithm) {
+            Ok(c) => c,
+            Err(_) => {
+                println!("⚠️  Skipping {:?} because it's not available.", algorithm);
+                return;
+            }
+        };
         let key = cipher.key_gen();
         let nonce = cipher.nonce_gen();
         let cipher = cipher.set_key(&key).unwrap();
@@ -171,7 +208,21 @@ fn key_gen_self_test() {
         };
         assert_eq!(msg[..], msg_decrypted[..]);
     }
-    run(Mode::Aes128Gcm);
-    run(Mode::Aes256Gcm);
+    if hacl_aes_available() {
+        run(Mode::Aes128Gcm);
+        run(Mode::Aes256Gcm);
+    } else {
+        println!("⚠️  AES NOT AVAILABLE ON THIS PLATFORM!")
+    }
     run(Mode::Chacha20Poly1305);
 }
+
+fn hacl_aes_available() -> bool {
+    unsafe {
+        EverCrypt_AutoConfig2_has_pclmulqdq()
+            && EverCrypt_AutoConfig2_has_avx()
+            && EverCrypt_AutoConfig2_has_sse()
+            && EverCrypt_AutoConfig2_has_movbe()
+            && EverCrypt_AutoConfig2_has_aesni()
+    }
+}
